[Setting name="Show output window", category="General"]
bool showOutputWindow = false;

[Setting name="Enable plugin", category="General"]
bool isEnabled = false;

int outputWindowFlags = UI::WindowFlags::NoResize | UI::WindowFlags::NoCollapse | UI::WindowFlags::AlwaysAutoResize | UI::WindowFlags::NoFocusOnAppearing;

array<double> preciseTriggerTimesPool;
array<double> preciseFinishTimeCandidates;
double preciseFinishTimeFinal = 0.0;

Dev::HookInfo@ PTT   = null;
Dev::HookInfo@ SPFTC = null;
Dev::HookInfo@ EPFTC = null;
bool isEngaged = false;


void EngagePlugin()
{
  if(!isEngaged)
  {
#if MP41 && MANIA64
    @PTT   = Dev::Hook(Dev::BaseAddress() + 0x73D24D, 8, "PreciseTriggerTimeHook");
    @SPFTC = Dev::Hook(Dev::BaseAddress() + 0xEB449B, 2, "SavePreciseFinishTimeCandidatesHook");
    @EPFTC = Dev::Hook(Dev::BaseAddress() + 0xEC0005, 2, "EvaluatePreciseFinishTimeCandidatesHook");
    isEngaged = true;
#elif TURBO
    @PTT   = Dev::Hook(Dev::BaseAddress() + 0xA857EC, 3, "PreciseTriggerTimeHook");
    @SPFTC = Dev::Hook(Dev::BaseAddress() + 0xA85880, 2, "SavePreciseFinishTimeCandidatesHook");
    @EPFTC = Dev::Hook(Dev::BaseAddress() + 0xA7BC60, 1, "EvaluatePreciseFinishTimeCandidatesHook");
	isEngaged = true;
#endif
    if(isEngaged)
    {
      print("Enabled Nanosecond functionality");
    }
    else
    {
      warn("Failed to enable Nanosecond functionality, this game version is not supported");
    }
  }
}

void DisengagePlugin()
{
  if(isEngaged)
  {
    Dev::Unhook(EPFTC);
    Dev::Unhook(SPFTC);
    Dev::Unhook(PTT);
	
	isEngaged = false;
	
	print("Disabled Nanosecond functionality");
  }
}


void OnSettingsChanged()
{
  if(isEnabled)
  {
    EngagePlugin();
  }
  else
  {
    DisengagePlugin();
  }
}


void RenderMenu()
{
  if(UI::MenuItem("\\$ff0" + Icons::ClockO + "\\$z Show Nanoseconds", "", isEnabled, true))
  {
    isEnabled = !isEnabled;
	
	OnSettingsChanged();
	
	showOutputWindow = isEnabled;
  }
}

void RenderInterface()
{
  if(showOutputWindow) {
    if(UI::Begin("Nanoseconds", showOutputWindow, outputWindowFlags)) {
      UI::Text("Race time: " + FormatPreciseTime(preciseFinishTimeFinal));
    }
    UI::End();
  }
}


#if MP41 && MANIA64
void PreciseTriggerTimeHook(uint64 rdx)
{
  double preciseTriggerTime = double(rdx) / 262144.0;
#elif TURBO
void PreciseTriggerTimeHook(uint32 eax, uint32 ecx, uint32 edi)
{
  double preciseTriggerTime = double(ecx - eax + (uint64(edi) << 0x20)) / 1000000.0;
#else
// Non-supported version, this will never be executed
void PreciseTriggerTimeHook()
{
  double preciseTriggerTime = double(0);
#endif
  preciseTriggerTimesPool.InsertLast(preciseTriggerTime);
}

void SavePreciseFinishTimeCandidatesHook()
{
  preciseFinishTimeCandidates = preciseTriggerTimesPool;
  preciseTriggerTimesPool = array<double>();
}

#if MP41 && MANIA64
void EvaluatePreciseFinishTimeCandidatesHook(uint64 rcx)
{
  uint64 addr = rcx + 0x58;
#elif TURBO
void EvaluatePreciseFinishTimeCandidatesHook(uint32 eax)
{
  uint32 addr = eax + 0x164;
#else
// Non-supported version, this will never be executed
void EvaluatePreciseFinishTimeCandidatesHook()
{
  uint32 addr = 0x0;
#endif
  double preciseFinishTime = 1e+300;
  for(uint i = 0; i < preciseFinishTimeCandidates.Length; i++)
  {
    if(preciseFinishTimeCandidates[i] < preciseFinishTime)
	{
	  preciseFinishTime = preciseFinishTimeCandidates[i];
	}
  }
  
  preciseFinishTimeCandidates = array<double>();
  
  int32 timeOffset = Dev::ReadInt32(addr);
  
  preciseFinishTimeFinal = (preciseFinishTime - timeOffset) / 1000.0;
  
  showOutputWindow = true;
  
  print("Race Time: " + FormatPreciseTime(preciseFinishTimeFinal));
}


string FormatPreciseTime(double time)
{
  int32 truncTime = int(time * 1000);
  return Time::Format(truncTime) + Text::Format("%06d", int((time * 1000 - truncTime) * 1e+6));
}


void Main()
{
  if(isEnabled)
  {
    EngagePlugin();
  }
}

void OnEnabled()
{
  Main();
}

void OnDisabled()
{
  DisengagePlugin();
}

void OnDestroyed()
{
  OnDisabled();
}